#!/usr/bin/env python
"""CDE checks for models from MDFs."""

from __future__ import annotations

import logging
import re
from json import JSONDecodeError
from typing import TYPE_CHECKING

import click
import requests
import stamina
from bento_mdf.mdf.reader import MDFReader

if TYPE_CHECKING:
    from bento_meta.model import Model
    from bento_meta.objects import Property
from packaging.version import Version

DEFAULT_TIMEOUT = 60
DEFAULT_RETRIES = 3
DEFAULT_RETRY_DELAY = 1.0
MAX_LINE_WIDTH = 80
INDENT = "    "

logger = logging.getLogger(__name__)


def format_value_list(values: list[str], indent_level: int = 0) -> str:
    """Format a list of values with proper wrapping and indentation."""
    if not values:
        return ""

    indent = INDENT * indent_level
    lines = []
    current_line = []
    current_length = len(indent)

    for i, value in enumerate(values):
        # Add quotes around the value and comma if not last
        formatted_value = f"'{value}'"
        if i < len(values) - 1:
            formatted_value += ","

        # Check if adding this value would exceed line width
        if current_line and current_length + len(formatted_value) + 1 > MAX_LINE_WIDTH:
            # Join current line and add to lines
            lines.append(indent + " ".join(current_line))
            current_line = [formatted_value]
            current_length = len(indent) + len(formatted_value)
        else:
            current_line.append(formatted_value)
            current_length += (
                len(formatted_value) + 1 if current_line else len(formatted_value)
            )

    # Add the last line
    if current_line:
        lines.append(indent + " ".join(current_line))

    return "\n".join(lines)


def print_formatted_output(message: str, indent_level: int = 0) -> None:
    """Print message with proper indentation."""
    indent = INDENT * indent_level
    print(f"{indent}{message}")


def get_pvs_from_cde_json(json: dict) -> list[str]:
    """Get CDE PVs from a CDE JSON response."""
    try:
        vs = []
        de = json.get("DataElement", {})
        if not de:
            return vs
        cde_pvs = de.get("ValueDomain", {}).get("PermissibleValues", [])
        if not cde_pvs:
            return vs
        vs.extend([pv["value"] for pv in cde_pvs])
    except Exception as e:
        msg = f"Exception occurred when getting value set from JSON: {e}"
        logger.exception(msg)
        return []
    else:
        return vs


def check_cde_version(
    prop: Property,
    cde_id: str,
    current_version: str | None,
) -> None:
    """Check for new CDE versions for CDE-annotated properties."""
    url = f"https://cadsrapi.cancer.gov/rad/NCIAPI/1.0/api/DataElement/{cde_id}"
    headers = {"accept": "application/json"}
    try:
        response = requests.get(
            url,
            timeout=DEFAULT_TIMEOUT,
            headers=headers,
        )
        response.raise_for_status()
        de = response.json().get("DataElement", {})
        if not de:
            print(f"{INDENT}No response from {url} for {cde_id} of {prop.handle}")
            return
        latest_version = de.get("version", "")
        if not current_version:
            print(
                f"{INDENT}No CDE version specified for '{prop.handle}'. "
                f"Latest version is '{latest_version}'",
            )
        elif Version(str(current_version)) != Version(latest_version):
            print(
                f"{INDENT}CDE annotation for prop: '{prop.handle}' with id: '{cde_id}' "
                f"and version: '{current_version}' doesn't match "
                f"latest version: '{latest_version}'",
            )
    except JSONDecodeError:
        msg = (
            f"Failed to parse JSON response for entity {prop.handle} "
            f"annotated with CDE {cde_id}v{current_version}"
        )
        logger.exception(msg)
        return
    except requests.HTTPError:
        msg = f"HTTP error fetching CDE {cde_id}v{current_version} for entity {prop.handle}"
        logger.exception(msg)
        return


@stamina.retry(on=requests.RequestException, attempts=DEFAULT_RETRIES)
def fetch_cde_pvs(
    prop: Property,
    cde_id: str,
    cde_version: str | None = None,
) -> list[str]:
    """Fetch CDE PVs for a given CDE version."""
    version_str = (
        f"?version={cde_version}"
        if cde_version and re.match(r"^v?\d{1,3}(\.\d{1,3}){0,2}$", cde_version)
        else ""
    )
    cde_param = f"{cde_id}{version_str}"
    url = f"https://cadsrapi.cancer.gov/rad/NCIAPI/1.0/api/DataElement/{cde_param}"
    headers = {"accept": "application/json"}
    try:
        response = requests.get(url, timeout=DEFAULT_TIMEOUT, headers=headers)
        response.raise_for_status()
        if not response:
            print(f"{INDENT}No response from {url}")
            return []
        json_response = response.json()
        value_set = get_pvs_from_cde_json(json_response)
    except JSONDecodeError as e:
        msg = f"Failed to parse JSON response for entity {prop.handle}: {e}\nurl: {url}"
        logger.exception(msg)
        return []
    except requests.HTTPError:
        msg = f"HTTP error fetching CDE {cde_id}v{version_str} for entity {prop.handle}"
        logger.exception(msg)
        return []
    else:
        return value_set


def check_enum_against_cde_pvs(model: Model, output_format: str = "text") -> None:
    """Compare enumerated value sets to CDE PVs for CDE-annotated properties."""
    already_checked = set()
    results = []  # For table format

    # Add section header
    if output_format == "text":
        print("\n##################################")
        print("#                                #")
        print("#        CDE Validation          #")
        print("#                                #")
        print("##################################")
        print("\nChecking CDE annotations against caDSR for CDE-annotated properties:")
        print("----------\n")

    for prop in model.props.values():
        prop_annotations = prop.annotations
        if not prop_annotations:
            continue
        annotations = prop_annotations.values()
        if not any("cadsr" in t.origin_name.lower() for t in annotations):
            continue
        for term in annotations:
            if "cadsr" not in term.origin_name.lower():
                continue
            cde_id = term.origin_id
            cde_version = term.origin_version
        key = (prop.handle, cde_id, cde_version)
        if key in already_checked:
            continue

        logger.info(
            f"Checking CDE '{cde_id}' 'v{cde_version}' for prop '{prop.handle}'",
        )
        already_checked.add(key)
        check_cde_version(prop, cde_id, cde_version)
        cde_pvs = fetch_cde_pvs(prop, cde_id, cde_version)
        if not cde_pvs:
            continue
        model_terms = prop.values
        if not model_terms:
            continue

        pvs_not_in_enum = [pv for pv in cde_pvs if pv not in model_terms]
        if pvs_not_in_enum:
            if output_format == "text":
                print(
                    f"{INDENT}CDE '{cde_id}' 'v{cde_version}' has {len(pvs_not_in_enum)} "
                    f"PVs not in enumerated value set for '{prop.handle}':",
                )
                print(format_value_list(sorted(pvs_not_in_enum)))
            else:
                results.append(
                    {
                        "Property": prop.handle,
                        "CDE ID": cde_id,
                        "CDE Version": cde_version or "",
                        "Issue Type": "CDE PVs not in Model",
                        "Count": len(pvs_not_in_enum),
                        "Values": ", ".join(sorted(pvs_not_in_enum)),
                    },
                )

        vals_not_in_cde_pvs = [v for v in model_terms if v not in cde_pvs]
        if vals_not_in_cde_pvs:
            if output_format == "text":
                print(
                    f"{INDENT}Enumerated value set for '{prop.handle}' has "
                    f"{len(vals_not_in_cde_pvs)} values not in CDE '{cde_id}' "
                    f"'v{cde_version}':",
                )
                print(format_value_list(sorted(vals_not_in_cde_pvs)))
            else:
                results.append(
                    {
                        "Property": prop.handle,
                        "CDE ID": cde_id,
                        "CDE Version": cde_version or "",
                        "Issue Type": "Model values not in CDE",
                        "Count": len(vals_not_in_cde_pvs),
                        "Values": ", ".join(sorted(vals_not_in_cde_pvs)),
                    },
                )

        already_checked.add(key)

    # If table format requested, print the table
    if output_format == "table" and results:
        print_table(results)


def print_table(results: list[dict]) -> None:
    """Print results in a simple table format."""
    # Group results by property
    property_groups = {}
    for result in results:
        prop = result["Property"]
        if prop not in property_groups:
            property_groups[prop] = []
        property_groups[prop].append(result)

    print("\n\tCDE Validation Results")
    print("\t----------")
    print(
        "\t+--------------------+-------------+----------+--------------------------+---------------+",
    )
    print(
        "\t| property           | cde_id      | version  | check                    | error count   |",
    )
    print(
        "\t+--------------------+-------------+----------+--------------------------+---------------+",
    )

    for prop, results_list in property_groups.items():
        for i, result in enumerate(results_list):
            if i == 0:
                prop_display = prop[:18] + ".." if len(prop) > 20 else prop
            else:
                prop_display = ""

            issue_display = (
                result["Issue Type"][:24] + ".."
                if len(result["Issue Type"]) > 26
                else result["Issue Type"]
            )

            print(
                f"\t| {prop_display:<18} | {result['CDE ID']:<11} | {result['CDE Version']:<8} | {issue_display:<24} | {result['Count']!s:<13} |",
            )

    print(
        "\t+--------------------+-------------+----------+--------------------------+---------------+",
    )

    # Print detailed values below the table
    print("\n\tDetailed error values:")
    print("\t----------")
    for prop, results_list in property_groups.items():
        if any(r["Count"] > 0 for r in results_list):
            print(f"\n\t{prop}:")
            for result in results_list:
                if result["Count"] > 0:
                    print(f"\t  {result['Issue Type']} ({result['Count']} values):")
                    # Get full list of values from the result
                    print(f"\t    {result['Values']}")


@click.command()
@click.option(
    "-f",
    "--mdf-files",
    help="MDF file(s) to compare. If you have multiple, use this for each file.",
    required=True,
    type=str,
    prompt=True,
    multiple=True,
)
@click.option(
    "--output_format",
    type=click.Choice(["text", "table"]),
    default="text",
    help="Output format: text (default) or table",
)
def main(mdf_files: str | list[str], output_format: str) -> None:
    """Compare enum from MDFs to CDE PVs from caDSR for CDE-annotated properties."""
    mdf = MDFReader(*mdf_files, raise_error=True)
    model = mdf.model
    check_enum_against_cde_pvs(model, output_format=output_format)


if __name__ == "__main__":
    main()  # type: ignore reportCallIssue



#!/usr/bin/env python
# from pdb import set_trace
import logging
from argparse import ArgumentParser, FileType
from sys import exit

from ..mdf import MDF
from ..validator import MDFValidator

ap = ArgumentParser(description="Validate MDF against JSONSchema")
ap.add_argument(
    "--schema",
    help="MDF JSONschema file",
    type=FileType("r"),
    dest="schema",
)
ap.add_argument(
    "--quiet",
    help="Suppress output; return only exit value",
    action="store_true",
    dest="quiet",
)
ap.add_argument(
    "mdf_files",
    nargs="+",
    metavar="mdf-file",
    type=FileType("r"),
    help="MDF yaml files for validation",
)
ap.add_argument("--log-file", help="Log file name")


def test(args, logger):
    retval = 0
    v = MDFValidator(args.schema, *args.mdf_files, logger=logger)
    if not v.load_and_validate_schema():
        retval += 1
    if not v.load_and_validate_yaml():
        retval += 1
    if not v.validate_instance_with_schema():
        retval += 1
    if not retval:
        for f in args.mdf_files:
            f.seek(0)
        if not MDF(
            *args.mdf_files,
            handle="test",
            logger=logger,
            ignore_enum_by_reference=True,
        ).create_model_success:
            retval += 1
    return retval


def do_test():
    args = ap.parse_args()
    logger = logging.getLogger("test-mdf")
    logger.setLevel(logging.DEBUG)
    fmt = logging.Formatter(fmt="%(asctime)s:%(name)s (%(levelname)s) - %(message)s")
    shdl = logging.StreamHandler()
    shdl.setLevel(logging.INFO)
    shdl.setFormatter(fmt)
    logger.addHandler(shdl)
    if args.quiet:
        shdl.setLevel(logging.CRITICAL)
    if args.log_file:
        fhdl = logging.FileHandler(args.log_file)
        fhdl.setLevel(logging.DEBUG)
        fhdl.setFormatter(fmt)
        logger.addHandler(fhdl)
    exit(test(args, logger))  # emit return val (0 = good) to os


if __name__ == "__main__":
    do_test()



#!/usr/bin/env python
"""Generate, view, and save model diff."""

from __future__ import annotations

import logging
import os
from pathlib import Path

import click
from bento_mdf.diff import diff_models
from bento_mdf.mdf.reader import MDFReader

logger = logging.getLogger("__name__")


@click.command()
@click.option(
    "--model_handle",
    required=True,
    type=str,
    prompt=True,
    help="CRDC Model Handle (e.g. 'GDC')",
)
@click.option(
    "--old_mdfs",
    required=True,
    type=str,
    prompt=True,
    multiple=True,
    help=(
        "Older version of MDF file(s). "
        "If MDF split into multiple files, provide url or path to each file."
    ),
)
@click.option(
    "--new_mdfs",
    required=True,
    type=str,
    prompt=True,
    multiple=True,
    help=(
        "Newer version of MDF file(s). "
        "If MDF split into multiple files, provide url or path to each file."
    ),
)
@click.option(
    "--old_version",
    required=False,
    type=str,
    help="Older version of MDF file(s) if not included in MDF",
)
@click.option(
    "--new_version",
    required=False,
    type=str,
    help="Newer version of MDF file(s)",
)
@click.option(
    "--output_path",
    required=False,
    type=str,
    prompt=False,
    help="File path for output diff. If not provided, diff will be save to cwd.",
)
@click.option(
    "--include_summary",
    required=False,
    type=bool,
    default=True,
    help="Include a summary of the diff in the result",
)
@click.option(
    "--objects_as_dicts",
    required=False,
    type=bool,
    default=False,
    help="Return dictionaries with entity attributes instead of bento_meta objects",
)
@click.option(
    "--summary_only",
    required=False,
    type=bool,
    default=False,
    help="Only include the diff summary in the result",
)
def main(  # noqa: PLR0913
    model_handle: str,
    old_mdfs: str | Path | list[str | Path],
    new_mdfs: str | Path | list[str | Path],
    old_version: str | None = None,
    new_version: str | None = None,
    *,
    include_summary: bool = True,
    objects_as_dicts: bool = True,
    output_path: Path | str | None,
    summary_only: bool = False,
) -> None:
    """Diff two versions of MDF files for a model."""
    old_mdf = MDFReader(*old_mdfs, handle=model_handle)
    new_mdf = MDFReader(*new_mdfs, handle=model_handle)

    if old_version:
        old_mdf.model.version = old_version
    if new_version:
        new_mdf.model.version = new_version

    summary = include_summary or summary_only

    diff = diff_models(
        mdl_a=old_mdf.model,
        mdl_b=new_mdf.model,
        objects_as_dicts=objects_as_dicts,
        include_summary=summary,
    )

    result = diff.get("summary") if summary_only else diff
    output_file_extension = "txt" if summary_only else "py"

    if output_path is None:
        output_path = (
            Path.cwd() / f"{model_handle}_diff_"
            f"{old_mdf.model.version}_{new_mdf.model.version}.{output_file_extension}"
        )
    dict_as_str = repr(result)
    with Path(output_path).open("w+", encoding="utf-8") as f:
        f.write(f"diff = {dict_as_str}")
    try:
        os.startfile(output_path)  # noqa: S606
    except OSError as e:
        msg = f"Error opening the file: {e}"
        logger.exception(msg)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Load model in MDF into an MDB."""

import sys

sys.path.insert(0, "..")
import argparse
import getpass
from sys import stderr

from bento_mdf.mdf import MDF
from bento_meta.mdb import (
    WriteableMDB,
    load_mdf,
    load_model_statements,
    make_nanoid,
)


def main() -> None:
    """Entry point for loading MDF into an MDB."""
    parser = argparse.ArgumentParser(description="Load model in MDF into an MDB")
    parser.add_argument(
        "files",
        nargs="*",
        metavar="MDF-FILE",
        help="MDF file(s)/url(s)",
    )
    parser.add_argument(
        "--commit",
        default="",
        help="commit SHA1 for MDF instance (if any)",
        required=True,
    )
    parser.add_argument("--handle", help="model handle")
    parser.add_argument("--user", help="MDB username")
    parser.add_argument("--passw", help="MDB password")
    parser.add_argument(
        "--bolt",
        metavar="BoltURL",
        help="MDB Bolt url endpoint (specify as 'bolt://...')",
    )
    parser.add_argument("--put", action="store_true", help="Load model to database")
    parser.add_argument(
        "--make-nanoids",
        action="store_true",
        help="Add new nanoids to graph nodes",
    )
    # example:
    # args = parser.parse_args([
    #     "https://raw.githubusercontent.com/CBIIT/icdc-model-tool/master/model-desc/icdc-model.yml",
    #     "https://raw.githubusercontent.com/CBIIT/icdc-model-tool/master/model-desc/icdc-model-props.yml",
    #     "--commit",
    #     "a4aa9a43b9ad2087638ceaeef50d4a22a4b9b959",
    #     "--handle",
    #     "ICDC",
    #     "--bolt",
    #     "bolt://localhost:7687",
    #     "--user",
    #     "neo4j",
    #     "--passw",
    #     getpass.getpass()
    #     ])

    args = parser.parse_args()

    if not args.files:
        parser.print_help(file=stderr)
        parser.exit(1)

    if args.put and not args.bolt:
        print("error: --bolt and --user args required to put to database", file=stderr)
        parser.exit(2)

    if args.put and not args.passw:
        args.passw = getpass.getpass()

    print("load model from MDFs", file=stderr)
    mdf = MDF(*args.files, handle=args.handle, _commit=args.commit, raise_error=True)
    model = mdf.model

    if args.put:
        print("Put model to DB", file=stderr)
        mdb = WriteableMDB(uri=args.bolt, user=args.user, password=args.passw)
        model.mdb = mdb
        load_mdf(mdf, model.mdb)
        if args.make_nanoids:
            print("Add nanoids to nodes", file=stderr)
            with mdb.driver.session() as s:
                result = s.run(
                    "match (n {_commit:$commit}) where not exists(n.nanoid) "
                    "with n limit 1 set n.nanoid=$nanoid return n",
                    {"commit": args.commit, "nanoid": make_nanoid()},
                )
                while result.peek():
                    result = s.run(
                        "match (n {_commit:$commit}) where not exists(n.nanoid) "
                        "with n limit 1 set n.nanoid=$nanoid return n",
                        {"commit": args.commit, "nanoid": make_nanoid()},
                    )
    else:
        for s in load_model_statements(model, args.commit):
            print(s)


if __name__ == "__main__":
    main()
